#!/bin/bash
#==============================================================================
# Copyright and license info is available in the LICENSE file included with
# the Server Deployment Package (SDP), and also available online:
# https://swarm.workshop.perforce.com/projects/perforce-software-sdp/view/main/LICENSE
#------------------------------------------------------------------------------

#==============================================================================
# Declarations and Environment

# Allow override of P4U_HOME, which is set only when testing P4U scripts.
export P4U_HOME=${P4U_HOME:-../main/Server/Unix/p4/common/bin}
export P4U_LIB=${P4U_LIB:-../main/Server/Unix/p4/common/lib}
export P4U_ENV=$P4U_LIB/p4u_env.sh
export P4U_LOG="/tmp/package_downloads.sdp.$(date +'%Y%m%d-%H%M%S').log"

# Environment isolation.  For stability and security reasons, prepend
# PATH to include dirs where known-good scripts exist.
# known/tested PATH and, by implication, executables on the PATH.
export PATH=$P4U_HOME:$PATH:~/bin:.
export P4CONFIG=${P4CONFIG:-.p4config}

[[ -r "$P4U_ENV" ]] || {
   echo -e "\nError: Cannot load environment from: $P4U_ENV\n\n"
   exit 1
}

declare BASH_LIBS=$P4U_ENV
BASH_LIBS+=" $P4U_LIB/libcore.sh"
BASH_LIBS+=" $P4U_LIB/libp4u.sh"

for bash_lib in $BASH_LIBS; do
   source $bash_lib
done

export VERBOSITY=3

declare Version=1.3.2
declare -i SilentMode=0
declare AppHome=${PWD%/*}
declare ExcludesCfg=$AppHome/tools/excludes.downloads.cfg
declare CommonFilesList=$AppHome/tools/common_files.txt

#==============================================================================
# Local Functions

#------------------------------------------------------------------------------
# Function: terminate
function terminate
{
   # Disable signal trapping.
   trap - EXIT SIGINT SIGTERM

   # Don't litter.
   cleanTrash

   vvmsg "$THISSCRIPT: EXITCODE: $OverallReturnStatus"

   # Stop logging.
   [[ "${P4U_LOG}" == off ]] || stoplog

   # With the trap removed, exit.
   exit $OverallReturnStatus
}

#------------------------------------------------------------------------------
# Function: usage (required function)
#
# Input:
# $1 - style, either -h (for short form) or -man (for man-page like format).
#------------------------------------------------------------------------------
function usage
{
   declare style=${1:--h}

   echo "USAGE for $THISSCRIPT v$Version:

$THISSCRIPT [-f] [-s] [-o] [-L <log>] [-si] [-v<n>] [-n] [-D]

or

$THISSCRIPT [-h|-man|-V]
"
   if [[ $style == -man ]]; then
      echo -e "
DESCRIPTION:
	Create an SDP distribution package files, *.tgz files for
	UNIX/Linux and *.zip for Windows, from the latest files
	in the SDP main branch.

	Prior to packaing, a variety of checks are done to ensure
	the latest files are synced, the user is logged in, no
	files are checked out in the main branch, etc.

	An exclusion config file defines files and directories
	to be excluded from packaging.  This exclusion config
	file has file name per line, which may include the
	'*' wild card.

	The exclusion config file can also specify directories to be
	excluded, which start with a 'D:' prefix.

	The exclusion config file is:
	$ExcludesCfg

OPTIONS:
 -f	Force creation of target tar/zip file, replacing an existing
	one if necessary.  The default behavior is to abort if
	the target package files already exists.

 -o	Proceed even if files are opened in Perforce and/or not synced
 	to the head revision.  This may be useful for testing purposes.

	By default, we abort if any files are checked out, or if files
	are not syncd to the head revision.  This option is not suitable
	for creating a distribution file that will be submitted.

 -v<n>	Set verbosity 1-5 (-v1 = quiet, -v5 = highest).

 -L <log>
	Specify the path to a log file, or the special value 'off' to disable
	logging.  By default, all output (stdout and stderr) goes to:
	$(dirname ${P4U_LOG}).

	NOTE: This script is self-logging.  That is, output displayed on the
	screen is simultaneously captured in the log file.  Do not run this
	script with redirection operators like '> log' or '2>&1', and do not
	use 'tee.'

-si	Operate silently.  All output (stdout and stderr) is redirected to
	the log only; no output appears on the terminal.  This cannot be
	used with '-L off'.
      
 -n	No-Op.  Prints commands instead of running them.

 -D     Set extreme debugging verbosity.

HELP OPTIONS:
 -h	Display short help message
 -man	Display man-style help message
 -V	Dispay version info for this script and its libraries.

FILES:
	Common Files List file, listing files common to both UNIX/Linux and
	Windows packages:
	$CommonFilesList

	Exclusions Config File, listing files to exclude from all packages:
	$ExcludesCfg

EXAMPLES:
	Typical Usage is with no arguments:
	cd $AppHome/tools
	./package_downloads.sh

	Thus must be run from a workspace that maps at least (and ideally
	only) the folder //guest/perforce_software/sdp/...

SEE ALSO:
"
   fi

   exit 1
}

#==============================================================================
# Command Line Processing

declare -i shiftArgs=0
declare SDPUnixVersion=
declare -i ForceTarCreation=0
declare -i AllowOpenedFiles=0
declare TarFile=

set +u
while [[ $# -gt 0 ]]; do
   case $1 in
      (-f) ForceTarCreation=1;;
      (-o) AllowOpenedFiles=1;;
      (-h) usage -h;;
      (-man) usage -man;;
      (-V) show_versions; exit 1;;
      (-v1) export VERBOSITY=1;;
      (-v2) export VERBOSITY=2;;
      (-v3) export VERBOSITY=3;;
      (-v4) export VERBOSITY=4;;
      (-v5) export VERBOSITY=5;;
      (-L) export P4U_LOG=$2; shiftArgs=1;;
      (-si) SilentMode=1;;
      (-n) export NO_OP=1;;
      (-D) set -x;; # Debug; use 'set -x' mode.
      (*) usageError "Unknown arg ($1).";;
   esac

   # Shift (modify $#) the appropriate number of times.
   shift; while [[ $shiftArgs -gt 0 ]]; do
      [[ $# -eq 0 ]] && usageError "Bad usage."
      shiftArgs=$shiftArgs-1
      shift
   done
done
set -u

#==============================================================================
# Command Line Verification

[[ $SilentMode -eq 1 && $P4U_LOG == off ]] && \
   usageError "Cannot use '-si' with '-L off'."

#==============================================================================
# Main Program

trap terminate EXIT SIGINT SIGTERM

declare -i OverallReturnStatus=0

if [[ "${P4U_LOG}" != off ]]; then
   touch ${P4U_LOG} || bail "Couldn't touch log file [${P4U_LOG}]."

   # Redirect stdout and stderr to a log file.
   if [[ $SilentMode -eq 0 ]]; then
      exec > >(tee ${P4U_LOG})
      exec 2>&1
   else
      exec >${P4U_LOG}
      exec 2>&1
   fi

   initlog
fi

declare TmpDir=$(mktemp -d)
declare SDPVersion=$(cat ../main/Version|cut -d '/' -f 3).$(cat ../main/Version|cut -d '/' -f 4|cut -d ' ' -f 1)
declare SDPVersion=$(echo $SDPVersion)
declare SDPUnixVersion=sdp.Unix.$SDPVersion
declare SDPWindowsVersion=sdp.Windows.$SDPVersion
declare DownloadsDir=$AppHome/downloads
declare TarFile=$DownloadsDir/$SDPUnixVersion.tgz
declare ZipFile=$DownloadsDir/$SDPWindowsVersion.zip
declare UnixWorkingDir=/tmp/$SDPUnixVersion
declare WindowsWorkingDir=/tmp/$SDPWindowsVersion
declare ChangeFile=$TmpDir/change.p4s
declare Changelist=
declare RelString=
declare RelDate=

runCmd "p4 -ztag -F %TicketExpiration% login -s" \
   "Checking login status." 0

if [[ -n "$CMDOUTPUT" ]]; then
   msg "Verified: Logged into Perforce."
else
   bail "Not logged into Perforce."
fi

runCmd "p4 -F %depotFile% -ztag opened ../main/..." \
   "Checking for opened files in main."

if [[ -n "$CMDOUTPUT" ]]; then
   if [[ $AllowOpenedFiles -eq 1 ]]; then
      msg "Proceeding despite opened files being detected, due to -o flag."
   else
      if [[ $CMDOUTPUT == *"//"* ]]; then
         bail "These files are checked out in main:\n$CMDOUTPUT\n\nAborting."
      else
         bail "Could not determine if files are checked out in main. Output was:\n$CMDOUTPUT\n\nAborting."
      fi
   fi
else
   msg "Verified: No files checked out in SDP main."
fi

runCmd "p4 -F %depotFile% -ztag sync -n ../..." \
   "Checking for unsynced SDP files."

if [[ -n "$CMDOUTPUT" ]]; then
   if [[ $AllowOpenedFiles -eq 1 ]]; then
      msg "Proceeding despite unsynced files, due to -o flag."
   else
      if [[ $CMDOUTPUT == *"//"* ]]; then
         bail "These files need to be synced:\n$CMDOUTPUT\n\nAborting."
      else
         bail "Could not determine if files need to be synced. Output was:\n$CMDOUTPUT\n\nAborting."
      fi
   fi
else
   msg "Verified: No unsynced SDP files."
fi

msg "AppHome=[$AppHome] TF=[$TarFile] ZF=[$ZipFile] UWD=[$UnixWorkingDir] WWD=[$WindowsWorkingDir]."

if [[ -r $TarFile ]]; then
   if [[ $ForceTarCreation -eq 1 ]]; then
      runCmd "/bin/rm -f $TarFile" \
         "Removing existing tar file [$TarFile] due to '-f'." ||\
            bail "Failed to remove existing tar file [$TarFile].  Aborting."
   else
      bail "Tar file already exists: [$TarFile].  Use '-f' to forcibly replace.  Aborting."
   fi
fi

if [[ -r $ZipFile ]]; then
   if [[ $ForceTarCreation -eq 1 ]]; then
      runCmd "/bin/rm -f $ZipFile" \
         "Removing existing zip file [$ZipFile] due to '-f'." ||\
            bail "Failed to remove existing zip file [$ZipFile].  Aborting."
   else
      bail "Zip file already exists: [$ZipFile].  Use '-f' to forcibly replace.  Aborting."
   fi
fi

msg "Creating empty working directories."
for workingDir in $UnixWorkingDir $WindowsWorkingDir; do
   if [[ -d "$workingDir" ]]; then
      runCmd "/bin/rm -rf $workingDir" \
         "Removing old working dir [$workingDir]." ||\
         bail "Failed to remove old working dir [$workingDir]. Aborting."
   fi

   runCmd "/bin/mkdir -p $workingDir" "Initializing empty working dir [$workingDir]." 0 ||\
      bail "Failed to create empty working dir [$workingDir]. Aborting."
done

if [[ ! -d "$DownloadsDir" ]]; then
   runCmd "/bin/mkdir -p $DownloadsDir" "Initializing empty downloads dir [$DownloadsDir]." 0 ||\
      bail "Failed to create downloads dir [$DownloadsDir]. Aborting."
fi

#------------------------------------------------------------------------------
# Unix *.tgz file.
SDPRoot="$UnixWorkingDir/sdp"

runCmd "/bin/cp -pr ${AppHome}/main $SDPRoot" \
   "Copying SDP package to Unix working dir." 0 ||\
   bail "The copy to the working dir failed.  Aborting."

cd "$SDPRoot" || bail "Failed to cd to sdp root in Working Dir [$SDPRoot]."

if [[ -r "$ExcludesCfg" ]]; then
   while read excludeFile; do
      [[ "$excludeFile" == "" || "$excludeFile" == "#"* ]] && continue
      if [[ "$excludeFile" == "D:"* ]]; then
         excludeDir=${excludeFile#D:}
         msg "/usr/bin/find $SDPRoot -type d -name \"$excludeDir\" -print -exec /bin/rm -r -f {} \\;"
         /usr/bin/find $SDPRoot -type d -name "$excludeDir" -print -exec /bin/rm -r -f {} \;
      else
         msg "/usr/bin/find $SDPRoot -type f -name \"$excludeFile\" -print -exec /bin/rm -f {} \\;"
         /usr/bin/find $SDPRoot -type f -name "$excludeFile" -print -exec /bin/rm -f {} \;
      fi
   done < $ExcludesCfg
fi

cd "$UnixWorkingDir" || bail "Failed to cd to Working Dir [$UnixWorkingDir]."

runCmd "tar -czf $TarFile sdp" ||\
   bail "Failed to create tar file [$TarFile]. Aborting."
#------------------------------------------------------------------------------
# Windows *.zip file.
SDPRoot="$WindowsWorkingDir/sdp"

runCmd "/bin/cp -pr ${AppHome}/main $SDPRoot" \
   "Copying SDP package to Windows working dir." 0 ||\
   bail "The copy to the working dir failed.  Aborting."

cd "$SDPRoot" || bail "Failed to cd to sdp root in Working Dir [$SDPRoot]."

if [[ -r "$ExcludesCfg" ]]; then
   while read excludeFile; do
      [[ "$excludeFile" == "" || "$excludeFile" == "#"* ]] && continue
      # For Windows, exclude Unix stuff in places where Windows would otherwise be excluded.
      excludeFile=$(echo $excludeFile|sed "s:Windows:Unix:g")
      if [[ "$excludeFile" == "D:"* ]]; then
         excludeDir=${excludeFile#D:}
         msg "/usr/bin/find $SDPRoot -type d -name \"$excludeDir\" -print -exec /bin/rm -r -f {} \\;"
         /usr/bin/find $SDPRoot -type d -name "$excludeDir" -print -exec /bin/rm -r -f {} \;
      else
         msg "/usr/bin/find $SDPRoot -type f -name \"$excludeFile\" -print -exec /bin/rm -f {} \\;"
         /usr/bin/find $SDPRoot -type f -name "$excludeFile" -print -exec /bin/rm -f {} \;
      fi
   done < $ExcludesCfg
fi

cd "$WindowsWorkingDir" || bail "Failed to cd to Working Dir [$WindowsWorkingDir]."

# Copy common files.
if [[ -r "$CommonFilesList" ]]; then
   while read srcFileSpec; do
      [[ "$srcFileSpec" == "" || "$srcFileSpec" == "#"* ]] && continue
      # Prefix Unix working dir to relative path specified in the Common Files List file.
      srcFileSpec=$UnixWorkingDir/sdp/$srcFileSpec
      tgtDir=$(echo "$srcFileSpec" | sed "s:Unix:Windows:g")
      msg "SFS=[$srcFileSpec] TD=[$tgtDir]"
      tgtDir=${tgtDir%\*}
      msg "TD2=[$tgtDir]"
      tgtDir=${tgtDir%\/*}
      msg "TD3=[$tgtDir]"
      for srcFile in $srcFileSpec; do
         msg "Copying [$srcFile] to [$tgtDir]."
         if [[ -d "$srcFile" ]]; then
            /bin/cp -pr "$srcFile" "$tgtDir" ||\
               bail "Failure performing copy of common subdirs. Aborting."
         else
            /bin/cp -p "$srcFile" "$tgtDir" ||\
               bail "Failure performing copy of common files. Aborting."
         fi
      done
   done < $CommonFilesList
else
   warnmsg "Skipping copy of common files from Unix to Windows, due to missing common files list [$CommonFilesList]."
   OverallReturnStatus=1
fi

# Convert ASCII text files to Windows line endings before packaging.
for textFile in $(/usr/bin/find . -type f -print | xargs file | egrep -i '(ascii|text)'| cut -d: -f1); do
   convertedTextFile=.tmp.converted
   /usr/bin/awk 'sub("$", "\r")' $textFile > $convertedTextFile
   /bin/mv -f "$convertedTextFile" "$textFile"
done

runCmd "zip -r $ZipFile sdp" ||\
   bail "Failed to create tar file [$ZipFile]. Aborting."

runCmd "tar -tzf $TarFile" \
   "Tar file [$TarFile] contains these elements:"

cd "$DownloadsDir" || bail "Failed to cd to downloads dir [$DownloadsDir]."

echo -e "Change: new\n
Description:\n\tPackaged SDP $SDPVersion." > $ChangeFile

p4 change -i < $ChangeFile > $OutFile 2>&1

egrep 'Change \d+ created.' $OutFile

if [[ $? -ne 0 ]]; then
   cat $OutFile
   bail "Failed to create pending changelist. Aborting."
fi

# Parse the generated changelist number from the output file.
Changelist=$(cat $OutFile)
Changelist=${Changelist#Change }
Changelist=${Changelist%% *}
Changelist=$(echo $Changelist)

runCmd "p4 rec -c $Changelist" "Reconciling in downloads dir [$DownloadsDir]." ||\
   bail" Failed to reconcile changes."

[[ -d "$TmpDir" ]] && /bin/rm -rf "$TmpDir"

if [[ $OverallReturnStatus -eq 0 ]]; then
   msg "${H}\nAll processing completed successfully.\n"
else
   msg "${H}\nProcessing completed, but with errors.  Scan above output carefully.\n" 
fi

# Illustrate using $SECONDS to display runtime of a script.
msg "That took about $(($SECONDS/3600)) hours $(($SECONDS%3600/60)) minutes $(($SECONDS%60)) seconds.\n"

# See the terminate() function, which is really where this script exits.
exit $OverallReturnStatus
